import { BehaviorSubject, Subject } from 'rxjs';
import { ClientRequest, ConnectionAttribute, ConnectionRequest, ConnectionState, Message, OutGoingInfo, ServerRequest, State } from '../interfaces/general.interface';
import { GrpcServiceMethod } from './grpc.service.method';
import { BufferService } from './buffer.service';
import * as dotenv from 'dotenv'
import * as _ from 'lodash'
dotenv.config()

export class ServerClientManager {
    private connectionAttributes: ConnectionAttribute[] = []
    private grpcService: GrpcServiceMethod = new GrpcServiceMethod()
    private defaultServerAttribute: ServerRequest = {
        name: 'Default - Server',
        serverUrl: "localhost:3000",
        connectionType: 'GRPC',
        messageToBePublishedFromApplication: new Subject<Message>()
    }
    private defaultClientAttribute: ClientRequest = {
        name: 'Default - Client',
        targetServer: "localhost:3001",
        connectionType: 'GRPC',
        messageToBeReceivedFromRemote: new Subject<Message>()
    }

    constructor() {
        // logic here
    }


    public async generateConnection(request: ConnectionRequest): Promise<any> {
        return new Promise(async (resolve, reject) => {
            let initialReport: ConnectionState
            let reportSubject: BehaviorSubject<ConnectionState>
            let retransmission: BufferService
            // let originalRequest = JSON.parse(JSON.stringify(request))
            let originalRequest = _.cloneDeep(request)
            let database: string
            let response: any = { message: `Fail to complete connection generation` }
            let statusChain: State = 0
            let connectionAttribute: ConnectionAttribute

            if (statusChain == 0) {
                if (!request.server) {
                    request.server = this.defaultServerAttribute
                }
                if (!request.client) {
                    request.client = this.defaultClientAttribute
                }
                if (request.database) {
                    database = request.database
                } else {
                    database = request.server.name + request.client.name
                }
                /* Inject retransmission here */
                initialReport = { status: 'BUFFER' }
                reportSubject = new BehaviorSubject(initialReport)
                retransmission = new BufferService(request.server.messageToBePublishedFromApplication, reportSubject, database)

                statusChain = 1
            }

            if (statusChain == 1) {
                // Connection Type checking
                if (request.server!.connectionType != request.client!.connectionType) {
                    console.log(`Connection Type DOES NOT MATCH!`)
                    statusChain = 0
                } else {
                    statusChain = 1
                }
            }

            if (statusChain == 1) {
                connectionAttribute = {
                    ConnectionID: {
                        local: request.server!.name + request.client!.name,
                        remote: request.client!.name + request.server!.name
                    },
                    outGoing: {
                        StreamID: request.server!.name,
                        PublisherID: request.server!.name,
                        SubscriberID: request.server!.name,
                        MessageToBePublished: retransmission!.getMessages(),
                        MessageToBeReceived: null
                    },
                    inComing: {
                        StreamID: request.client!.name,
                        PublisherID: request.client!.name,
                        SubscriberID: request.client!.name,
                        MessageToBePublished: null,
                        MessageToBeReceived: request.client!.messageToBeReceivedFromRemote
                    },
                    connectionStatus: reportSubject!
                }
                statusChain = 1
            }

            if (statusChain == 1) {
                await this.checkConnectionAttribute(connectionAttribute!).then((res) => {
                    if (res == true) {
                        console.log(`Connection<${connectionAttribute.ConnectionID.local}> already exists `)
                    }
                    if (res == false) {
                        this.connectionAttributes.push(connectionAttribute)
                        console.log(`Connection ${connectionAttribute.ConnectionID.local} registered...`)
                        response = {
                            message: "Channel Response",
                            requestedTo: originalRequest,
                            data: connectionAttribute
                        }
                    }
                    console.log(`There is now ${this.connectionAttributes.length} connection Attributes`)
                })
                statusChain = 1
            }

            if (statusChain == 1) {
                // This is default connection`
                if (!request.client!.connectionType) {
                    request.client!.connectionType = 'GRPC'
                }
                // For each connection type:
                if (request.client!.connectionType == 'GRPC') {
                    // this.grpcService.create(request, connectionAttribute, this.outGoingInfo)
                }
            }

            resolve(response);

        })
    }

    private async checkConnectionAttribute(connectionAttribute: ConnectionAttribute): Promise<boolean> {
        return new Promise((resolve) => {
            let result: boolean = this.connectionAttributes.some(connection =>
                connection.ConnectionID.local === connectionAttribute.ConnectionID.local
            );
            console.log(`Checking ${connectionAttribute.ConnectionID.local} and returns ${result}`);
            resolve(result);
        });
    }


}

